/*
 * index.c
 *
 * By Ross Ridge
 * Public Domain
 *
 * Creates an index of EDICT for lookup.
 *
 */

#include "defs.h"
#include "edict.h"

static char const sccs_id[] = "@(#) myjis index.c 1.4 99/01/15 19:18:39";

#define LIST_SIZE	120000
#define MAX_KEY_LEN	512

struct ent {
	wide_char *wkey;
	char *key;
	long offset;
};

int lineno;
struct ent *list;
size_t list_len;
size_t list_pos;

static int
add_ent(FILE *f, int endc, long offset, long *offsetp, int *lastc) {
	int c;
	char *s;
	wide_char *wkey;
	char buf[MAX_KEY_LEN+1];

	c = getc(f);
	++*offsetp;
	if (c == EOF) {
		*lastc = c;
		return 0;
	}

	s = buf;
	while(c != endc) {
		if (s == buf + MAX_KEY_LEN) {
			errno = EINVAL;
			return -1;
		}
		*s++ = c;
		c = getc(f);
		++*offsetp;
		if (c == EOF) {
			errno = EINVAL;
			return -1;
		}
	}
	*s = '\0';
	if (list_pos == list_len) {
#if 0
		fputc('*', stderr);
#endif
		list_len *= 2;
		list = xrealloc(list, list_len * sizeof(struct ent));
	}

	wkey = jis_to_wide_dup(buf);
	if (wkey == NULL) {
		return -1;
	}
	wide_to_hiragana_str(wkey);
	wide_to_hiragana_vowel_str(wkey);

	list[list_pos].key = xstrdup(buf);
	list[list_pos].wkey = wkey;
	list[list_pos].offset = offset;
	list_pos++;

	return c;
}

static int
cmp_ent(void const *v1, void const *v2) {
	struct ent const *p1 = v1;
	struct ent const *p2 = v2;
	return wide_cmp(p1->wkey, p2->wkey);
}

static int
cmp_ent_progress(void const *v1, void const *v2) {
	struct ent const *p1 = v1;
	struct ent const *p2 = v2;
	static int count;
	if (count++ % 50000 == 0) {
		putc('.', stderr);
		fflush(stderr);
	}
	return wide_cmp(p1->wkey, p2->wkey);
}

static int
make_index(FILE *out, FILE *in, int progress) {
	long offset = 0;
	long offset_save;
	int c;
	size_t i;
	int ret;

	list_pos = 0;
	list_len = LIST_SIZE;
	list = xmalloc(sizeof(struct ent) * list_len);

	ret = -1;

	while(1) {
		if (progress) {
			lineno++;
			if (lineno % 1000 == 0) {
				putc('.', stderr);
				fflush(stderr);
			}
		}

		offset_save = offset;
		if (add_ent(in, ' ', offset_save, &offset, &c) == -1) {
			break;
		}
		if (c == EOF) {
			ret = 0;
			break;
		}

		c = getc(in);
		offset++;
		if (c == '[') {	
			if (add_ent(in, ']', offset_save, &offset, &c) == -1) {
				break;
			}
		}
		while(c != EOF && c != '\n') {
			c = getc(in);
			offset++;
		}
		if (c == EOF) {
			errno = EINVAL;
			break;
		}
	}

	if (ret == 0) {
		if (progress) {
			fprintf(stderr, "\nentries: %d\nsorting", list_pos);
			fflush(stderr);
			qsort(list, list_pos, sizeof(struct ent),
			      cmp_ent_progress);
			putc('\n', stderr);
		} else {
			qsort(list, list_pos, sizeof(struct ent), cmp_ent);
		}

		for(i = 0; i < list_pos; i++) {
			if (fprintf(out, "%s %lx\n",
				    list[i].key, list[i].offset) < 0) {
				ret = -1;
				break;
			}
		}
	}

	for(i = 0; i < list_pos; i++) {
		free(list[i].key);
		free(list[i].wkey);
	}
	free(list);
	
	return ret;
}

int
edict_make_index(char const *index, char const *file, int progress) {
	FILE *in, *out;
	enum jis_locale_ctypes oldloc;
	int ret;

	in = fopen(file, "rb");
	if (in == NULL) {
		return -1;
	}

	out = fopen(index, "wb");
	if (out == NULL) {
		fclose(in);
		return -1;
	}

	oldloc = jis_set_locale_ctype(JIS_LOCALE_QUERY);
	jis_set_locale_ctype(JIS_LOCALE_EUC);
	
	ret = make_index(out, in, progress);

	jis_set_locale_ctype(oldloc);

	if (fclose(in) == EOF) {
		ret = -1;
	}

	if (fclose(out) == EOF) {
		ret = -1;
	}

	return ret;
}
