/*
 * kanjidic.h
 *
 * By Ross Ridge
 * Public Domain
 *
 * @(#) myjis kanjidic.h 1.1 97/02/21 01:44:09
 */

#ifndef __JIS_KANJIDIC_H__
#define __JIS_KANJIDIC_H__

#include "list.h"
#include "kindex.h"

class skip_code {
	unsigned short code;

public:
	skip_code(char const *s);
	skip_code(unsigned short n = 0): code(n) {}

	unsigned short
	packed() const {
		return code;
	}

#if 0
	operator unsigned short() const {
		return packed();
	}
#endif

	int
	operator >(skip_code sc) const {
		return code > sc.code;
	}

	int
	operator ==(skip_code sc) const {
		return code == sc.code;
	}

	char *unpacked() const;
};

typedef wide_char *reading;

class qcompare<reading> {
public:
	static int both_on_kun;
	static int fn(reading const &p1, reading const &p2);
};

// int qcompare<reading>::fn(reading const &s1, reading const &s2);

class kanjidict {
	FILE *f;

public:
	struct quick_entry {
		long offset;
		wide_char character;
		short score;

		quick_entry() {}
		quick_entry(long o, wide_char w, short s)
			: offset(o), character(w), score(s) {}
	};

private:
	list<quick_entry> quick_table;
	index<skip_code> skip_index;
	list<list<int>/**/> stroke_index;
	list<int> radical_index[214];
	auto_free_index<reading> reading_index;
	auto_free_index<reading> name_reading_index;

	static list<quick_entry> const *qtable;
	static int compare_by_score(void const *p1, void const *p2);

public:
	struct entry {
		short line;
		wide_char character;
		unsigned char radical, classic_radical;
		short frequency;
		char grade;
		char strokes;
		short halpern, nelson;
		skip_code skip;
		unsigned short unicode;
		char *spahn_hadamitzky;
		char *four_corner;
		struct {
			long index;
			short page;
			char vol;
		} morohashi;
		short eindex, kindex, lindex, oindex;
		auto_free_list<char *> korean;
		auto_free_list<char *> pinyin;
		auto_free_list<char *> xrefs;
		list<skip_code> bad_skip;
		auto_free_list<wide_char *> readings;
		auto_free_list<wide_char *> name_readings;
		auto_free_list<char *> translations;

		entry();
		int convert_line(char const *line, int lineno);
		~entry();
	};

	kanjidict(): f(NULL) {}

	int open(char const *file);

	int make_index();
	int read_index(char const *ifile);
	int write_index(char const *ifile) const;

	int length() const { return quick_table.length(); }

	wide_char
	operator [](int n) const {
		return quick_table[n].character;
	}

	int
	search_skip(skip_code key, list<int> &l) const {
		return skip_index.search(key, l);
	}

	list<int> const &
	search_stroke(char key) const {
		return stroke_index[key - 1];
	}

	list<int> const &
	search_radical(unsigned char key) const {
		assert(key > 0 && key <= 214);
		return radical_index[key - 1];
	}

	int
	search_reading(wide_char const *key, list<int> &l, int both) const {
		qcompare<reading>::both_on_kun = both;
		return reading_index.search((wide_char *) key, l);
	}

	int
	search_name_reading(wide_char const *key, list<int> &l) const {
		qcompare<reading>::both_on_kun = 1;
		return name_reading_index.search((wide_char *) key, l);
	}

	void
	sort_by_score(list<int> &lst) const {
		qtable = &quick_table;
		lst.qsort(compare_by_score);
	}

	void print_quick_table() const;
	void print_stroke_index() const;
	void print_radical_index() const;
	void print_reading_index() const;
	void print_skip_index() const;

	~kanjidict() {
		if (f != NULL) {
			fclose(f);
		}
	}
};

#endif /* __JIS_KANJIDIC_H__ */

/* Local variables: */
/* mode: c++ */
/* End: */
