/*
 * kd_entry.c
 *
 * By Ross Ridge
 * Public Domain
 *
 */

#include "defs.h"

#include "kanjidic.h"

#include <ctype.h>

#ifdef CONFIG_SCCS_IDS
static char const sccs_id[] = "@(#) myjis kd_entry.c 1.1 97/02/21 01:44:02";
#endif

kanjidict::entry::entry()
	: line(0), radical(0), classic_radical(0), frequency(0),
	  grade(0), strokes(0), halpern(0), nelson(0), unicode(0), 
	  spahn_hadamitzky(NULL), four_corner(NULL),
	  eindex(0), kindex(0), lindex(0), oindex(0) {
	morohashi.index = morohashi.page = morohashi.vol = 0;
}

static char *
make_string(char const *s, char delimiter = ' ') {
	if (s == NULL) {
		errno = EINVAL;
		return NULL;
	}

	char *end = strchr(s, delimiter);
	if (end == NULL) {
		errno = EINVAL;
		return NULL;
	}

	size_t len = end - s;
	char *str = (char *) malloc(len + 1);
	if (str == NULL) {
		return NULL;
	}
	memcpy(str, s, len);
	str[len] = '\0';

	return str;
}

static int
add_string(list<char *> &lst, char const *pos, char delimiter = ' ') {
	char *str = make_string(pos, delimiter);
	if (str == NULL) {
		return 1;
	}
	return lst.add(str);
}

static int
add_reading(list<wide_char *> &lst, char const *pos) {
	char *str = make_string(pos, ' ');
	if (str == NULL) {
		return 1;
	}

	enum jis_locale_ctypes old_loc 
		= jis_set_locale_ctype(JIS_LOCALE_QUERY);
	jis_set_locale_ctype(JIS_LOCALE_EUC);

	wide_char *wstr = jis_to_wide_dup(str);
	int save_errno = errno;

	jis_set_locale_ctype(old_loc);

	free(str);
	if (wstr == NULL) {
		errno = save_errno;
		return 1;
	}
	return lst.add(wstr);
}

int
kanjidict::entry::convert_line(char const *s, int lineno) {
	line = lineno;

#if 0
	/*
	 * Could (should?) zero out everything here.
	 */
	radical = classic_radical = 0;
	frequency = 0;
	grade_level = 0;
	strokes = 0;
	halpern = nelson = 0;
	unicode = 0;
	morohashi.index = morohashi.page = moroshashi.vol = 0;
	eindex = kindex = lindex = oindex = 0;
#endif
	
	char const *pos = s;
	int name_reading = 0;
	
	enum jis_locale_ctypes old_loc 
		= jis_set_locale_ctype(JIS_LOCALE_QUERY);
	jis_set_locale_ctype(JIS_LOCALE_EUC);

	jis_state state = __jis_initial_state;
	size_t n = jis_to_wide(&character, pos, strlen(pos), &state);
	int save_errno = errno;
	jis_set_locale_ctype(old_loc);

	if (n == (size_t) -1 || n == 0) {
		errno = save_errno;
		return 1;
	}

	pos += n;

	while(1) {
		pos = strchr(pos, ' ');
		if (pos == NULL) {
			return 0;
		}
		pos++;
#if 0
		printf("@@@: *pos = '%c'\n", *pos);
#endif
		switch(*pos) {
		case 'B':
			radical = atoi(pos + 1);
			break;

		case 'C':
			classic_radical = atoi(pos + 1);
			break;

		case 'F':
			frequency = atoi(pos + 1);
			break;

		case 'G':
			grade = atoi(pos + 1);
			break;

		case 'H':
			halpern = atoi(pos + 1);
			break;

		case 'N':
			nelson = atoi(pos + 1);
			break;

		case 'P':
			skip = skip_code(pos + 1);
			break;

		case 'S':
			strokes = atoi(pos + 1);
			break;

		case 'U':
			unicode = strtoul(pos + 1, NULL, 16);
			break;

		case 'I':
			spahn_hadamitzky = make_string(pos + 1, ' ');
			if (spahn_hadamitzky == NULL) {
				return 1;
			}
			break;

		case 'Q':
			four_corner = make_string(pos + 1, ' ');
			if (four_corner == NULL) {
				return 1;
			}
			break;

		case 'M':
			if (pos[1] == 'N') {
				morohashi.index = strtol(pos + 2, NULL, 10);
			} else if (pos[1] == 'P') {
				char *dot;
				morohashi.vol = strtol(pos + 2, &dot, 10);
				if (dot == NULL || *dot != '.') {
					return 1;
				}
				morohashi.page = strtol(dot + 1, NULL, 10);
			} else {
				errno = EINVAL;
				return 1;
			}
			break;

		case 'E':
			eindex = atoi(pos + 1);
			break;

		case 'K':
			kindex = atoi(pos + 1);
			break;

		case 'L':
			lindex = atoi(pos + 1);
			break;

		case 'O':
			oindex = atoi(pos + 1);
			break;

 		case 'W':
			if (add_string(korean, pos + 1)) {
				return 1;
			}
			break;

		case 'Y':
			if (add_string(pinyin, pos + 1)) {
				return 1;
			}
			break;

		case 'X':
			if (add_string(xrefs, pos + 1)) {
				return 1;
			}
			break;

		case 'Z':
			if (pos[1] != '\0' && pos[2] == 'P') {
				if (bad_skip.add(skip_code(pos + 3))) {
					return 1;
				}
			}
			break;

		case 'T':
			name_reading = 1;
			break;

		case '{':
#if 0
			printf("@@@: pos = %s\n", pos);
#endif
			if (add_string(translations, pos + 1, '}')) {
				return 1;
			}
			pos = strchr(pos, '}');
			if (pos == NULL) {
				printf("@@@: ack!\n");
				errno = EINVAL;
				return 1;
			}
			break;

		case ' ':
			break;

		case '\n':
			return 0;

		default:
			if (isalnum((unsigned char) pos[0])) {
				break;
			}

			if (name_reading) {
 				if (add_reading(name_readings, pos)) {
					return 1;
				}
			} else {
 				if (add_reading(readings, pos)) {
					return 1;
				}
			}
			break;
		}
	}
}

kanjidict::entry::~entry() {
	if (spahn_hadamitzky != NULL) {
		free(spahn_hadamitzky);
	}
	if (four_corner != NULL) {
		free(four_corner);
	}
}
