/*
 * pts.h
 *
 * By Ross Ridge
 * Public Domain
 *
 * @(#) myjis pts.h 1.1 00/05/15 14:04:34
 */

#ifndef __PTS_H__
#define __PTS_H__

#include "kanjidic.h"

#if 0
#include <ctype.h>
#include <signal.h>
#include <stdarg.h>
#include <unistd.h>
#include <stropts.h>
#include <sys/stream.h>
#include <sys/wait.h>
#endif

#include <sys/termios.h>
#include <sys/termio.h>

extern struct termios parent_termios;

extern FILE *flog;

#ifdef __GNUC__
void log(char const *fmt, ...) __attribute__((format(printf, 1, 2)));
#else
void log(char const *fmt, ...);
#endif

void log_buf(char const *buf, size_t len);
void log_list_int(list<int> const &lst);
void lerror(char const *msg);

class pt_link {
	int master, slave, ttyin, ttyout;

	enum { R2K_KANA_LEN = 10 };
	enum flush_arg { FLUSH_NONE, FLUSH_READ, FLUSH_WRITE, FLUSH_BOTH };

	class pending {
		struct elt {
			struct elt *next;
			char *data;
			ssize_t len;
		};


		elt *head;
		elt *tail;

		int poll_fd;

		void dequeue();

		int check_poll();

	protected:
		int in_fd, out_fd;

		virtual ssize_t read_data(char **len);
		virtual ssize_t write_data(char *data, ssize_t len);
		virtual int flush_read();
		virtual int flush_write();
		
	public:
		pending(): head(NULL), tail(NULL), poll_fd(-1),
			   in_fd(-1), out_fd(-1) {}

		void
		set_fds(int i, int o) {
			in_fd = i;
			out_fd = o;
		}

		int enqueue(char const *buf, ssize_t len);

		int set_poll();
		int clear_poll();
		int process();

		int flush(enum flush_arg arg = FLUSH_NONE);

		int is_pending() { return head != NULL; }

		virtual ~pending() { dequeue(); }
	};

	class pending_in_base: public pending {
	public:
		virtual int send_sig(int) { return 0; }

	};

	class pending_out;
	class pending_statline;

	class canon_in {
		char buf[512];
		signed char char_marker[512];
		int canon_pos;
		int backslash;
		int lnext;
		enum { NONE, COMMAND, KANJI, KANJI_INPUT } command_mode;
		enum imode { NORMAL, HIRAGANA, KATAKANA } input_mode;
		int ignore_bytes;
		int use_status_line;
		int recursive_edit;

		char const *status_line_prompt;
		char sl_msg_buf[1024];
		int sl_msg_pos;
		int sl_msg_offset;

		int translation_pos;
		const int r2k_flags;
		int r2k_kana_count;
		struct jis_r2k_kana const *r2k_kana[R2K_KANA_LEN];
		int r2k_last_substitute;

		struct termios slave_termios;
		struct termios old_slave_termios;
		struct termios master_termios;

		int icanon;
		int echo;

		/*
		 * EUC code set 1 must be one byte long and have a
		 * display width of one.
		 */
		static const int euc_byte_len[4];
		static const int euc_char_width[4];

		int
		euc_is_ss2(char c) {
			return c == (char) 142;
		}

		int
		euc_is_ss3(char c) {
			return c == (char) 143;
		}

		int
		euc_is_mb(char c) {
			return (unsigned char) c >= 160
				&& (unsigned char) c <= 255;
		}

		int
		euc_is_print(char c) {
			return euc_is_mb(c) || euc_is_ss2(c) || euc_is_ss3(c)
			       || (c >= 32 && (unsigned char) c <= 127);
		}

		class pending_fake_in: public pending_in_base {
			char *input;

			virtual ssize_t write_data(char *data, ssize_t len);

		public:
			pending_fake_in(): input(NULL) {}

			char *
			get_input() {
				char *ret = input;
				input = NULL;
				return ret;
			}

			~pending_fake_in() {
				if (input != NULL) {
					free(input);
				}
			}
		};
			
		pending &pending_canon_output;
		pending_in_base &pending_input;
		pending_out &pending_output;
		pending_statline &pending_status_line;

		kanjidict &kdict;
		canon_in *kinput;
		pending_fake_in fake_input;
		char kanji_command;
	       
		list<int> kanji_selection;

		int set_normal_mode();
		int set_prompt_mode();
		int set_long_prompt_mode();

		int flush_output();
		int status_line_echo(char c);
		int status_line_echo(char const *s);
		int write_echo(char c);
		int write_echo(char const *s);	
		int update_status_line(int force = 1);
		int print_mode(char const *s);
		int flush_prompt();
		int print_prompt(char c);
		int print_prompt(char const *s);
		int process_echo(char c);
		int process_interrupt(char c, int sig);
		int process_reprint();
		int process_erase(int echoe, int dest);

		int
		process_erase(int echoe) {
			return process_erase(echoe, canon_pos - 1);
		}

		int substitute_input(int pos, wide_char const *s,
				     int advance_pos);
		int roman_to_kana(char c);
		int process_character(char c, int translate = 1);
		int process_break(int pos);
		int process_break() { return process_break(canon_pos); }
		int process_command(char c);

		int select_skip(char const *input);
		int select_reading(char const *input, char cmd);
		int process_kanji_input(char c);
		int start_kanji_input(char const *prompt, char cmd);
		int merge_selection(list<int> const &select);
		int substitute_kanji_input(int n);
		int kanji_display_selection(int limit);
		int process_kanji_command(char c);

		int
		kanji_display_selection() {
			return kanji_display_selection(kdict.length());
		}

	public:
		canon_in(struct termios *parent_termios,
			 class pending &co, class pending_in_base &in,
			 class pending_out &out, class pending_statline &sl,
			 kanjidict &kdict, int use_statline,
			 int mode = NORMAL, int recursive = 0);

		int init();

		int process_canon_input(char c);
		int process_setattr(int cmd, struct termios *tio);
		int process_termio(int cmd, struct termio *otio);

		int
		raw() {
			return (slave_termios.c_lflag & (ISIG | ICANON)) == 0;
		}

		~canon_in();
	};

	class pending_in: public pending_in_base {
		class canon_in &canon_input;

		virtual ssize_t read_data(char **data);
		virtual int flush_read();

	public:
		pending_in(class canon_in &c)
			: canon_input(c) {}

		int
		send_sig(int sig) {
			/*
			 *  TIOCSIGNAL is undocumented.
			 */ 
			if (out_fd != -1 
			    && ioctl(out_fd, TIOCSIGNAL, sig) == -1) {
				lerror("TIOCSIGNAL");
				return -1;
			}
			return 0;
		}
	};

	class pending_out: public pending {
		canon_in &canon_input;
		
		enum {
			NORMAL,
			ESC,
			PARAM,
			FINAL
		} state;
			
		virtual ssize_t write_data(char *data, ssize_t len);
		virtual ssize_t read_data(char **len);
		virtual int flush_write();

		int process_ioctl(int cmd, char *data, size_t len);

	public:
		pending_out(class canon_in &c)
			: canon_input(c) {}

		int
		setattr(int cmd, struct termios *tio) {
			if (out_fd == -1) {
				return 0;
			}
			return ioctl(out_fd, cmd, tio);
		}

		int
		status_line_ok() {
			return !is_pending() && state == NORMAL;
		}
	};
	
	class pending_statline: public pending {
		pending_out &pending_output;
		
		virtual ssize_t write_data(char *data, ssize_t len);

	public:
		pending_statline(class pending_out &po)
			: pending_output(po) {}
	};

	pending_out pending_output;
	pending pending_canon_output;
	pending_statline pending_status_line;
	canon_in canon_input;
	pending_in pending_input;

	kanjidict kdict;

	int process_input();

	int process_canon_output();

public:
	pt_link(int ttyin, int ttyout, int use_status_line = 1);

	int open(char const *kdict, char const *index);
	int open_slave();
	int process(pid_t child, int *waitst);
	int close();

	~pt_link();
};

#endif /* __PTS_H__ */

/* Local variables: */
/* mode: c++ */
/* End: */
