/*
 * w_to_j.c
 *
 * By Ross Ridge
 * Public Domain
 *
 * Implementation of wide_to_jis().
 *
 */

#include "defs.h"

#ifdef CONFIG_SCCS_IDS
static char const sccs_id[] = "@(#) myjis w_to_j.c 1.11 98/06/01 22:28:33";
#endif

static struct {
	enum __jis_states state;
	char const *shift;
	size_t shiftlen;
	size_t bytelen;
} const jis_shifts[] = {
        { __JIS_ROMAN, "\033(J", 3, 1},
	{ __JIS_ASCII, "\033(B", 3, 1},
	{ __JIS_KATAKANA, "\033(I", 3, 1},
	{ __JIS_1983, "\033$B", 3, 2},
	{ __JIS_1990, "\033&@\033$B", 6, 2},
	{ __JIS_0212, "\033$(D", 4, 2},
};

static int (* const wide_to_kuten)(long wc, enum char_domain *tablep,
				   int *codep, int *code2p)
#ifdef __JIS_CONFIG_USE_UNICODE
		= __jis_unicode_to_kuten;
#else
		= __jis_euc_to_kuten;
#endif

#ifdef __JIS_CONFIG_USE_UNICODE

CONFIG_INLINE
static int
ascii_to_roman(int code) {
	if (code < 32) {
		if (code < 0) {
			return -1;
		}
		return code;
	} else if (code > 255) {
		return -1;
	} else if (code >= 127 && code <= 159) {
		return code;
	} else {
		int ret = __jis_wide_to_jis[0][code];
		if (ret >= 0 && ret <= 255) {
			return ret;
		}
		return -1;
	}
}

#ifdef CONFIG_EUC_ASCII

CONFIG_INLINE
static int
roman_to_ascii(int code) {
	if (code >= 0 && code <= 255) {
		unsigned ret = __jis_roman_to_wide[code];
		if (ret >= 0 && ret <= 255) {
			return ret;
		}
	}
	return -1;
}

#endif /* CONFIG_EUC_ASCII */

#else /* __JIS_CONFIG_USE_UNICODE */

CONFIG_INLINE
static int
ascii_to_roman(int code) {
	return code;
}

#ifdef CONFIG_EUC_ASCII

CONFIG_INLINE
static int
roman_to_ascii(int code) {
	return code;
}

#endif

#endif /* __JIS_CONFIG_USE_UNICODE */

#ifdef __JIS_CONFIG_HAVE_UNICODE

CONFIG_INLINE
static size_t
unicode_to_utf8(char *s, size_t len, long ucode) {
	unsigned long const masks[] = {
		0x0000007fL,
		0,
		0x000007ffL,
		0x0000ffffL,
		0x001fffffL,
		0x03ffffffL,
		0x7fffffffL,
	};
	int n;

	if ((ucode & ~0x0000007fL) == 0) {
		if (len < 1) {
			return 0;
		}
		if (s != NULL) {
			s[0] = ucode;
		}
		return 1;
	}

	n = 2;
	while((ucode & ~masks[n]) != 0) {
		if (n == 6) {
			errno = EILSEQ;
			return (size_t) -1;
		}
		n++;
	}
		
	if ((size_t) n > len) {
		return 0;
	}

#if 0
	printf("@@@ ucode = 0x%08lx, n = %d\n", ucode, n);
#endif

	if (s != NULL) {
		int i;
		static unsigned char const table[7] = {
			0x0, 0, 0xc0, 0xe0, 0xf0, 0xf8, 0xfc
		};
		
		for(i = n - 1; i > 0; i--) {
			s[i] = (char) ((ucode & 0x3f) | 0x80);
			ucode >>= 6;
		}
		s[0] = (char) (table[n] | (unsigned char) ucode);
	}

	return n;
}

#endif /* __JIS_CONFIG_HAVE_UNICODE */


size_t
wide_to_jis(char *s, size_t len, wide_char wc, jis_state *state) {
	int code, code2;
	enum char_domain table;
	enum __jis_states new_state;
	size_t bytelen, shiftlen;
	enum jis_locale_ctypes locale = __jis_get_locale(state);

	if (s == NULL) {
		len = JIS_MB_LEN_MAX;
		wc = (wide_char) 0;
	}

#ifdef __JIS_CONFIG_HAVE_UNICODE

	if (locale == JIS_LOCALE_UTF8) {
		long ucode = jis_wide_to_unicode(wc);
		if (ucode == -1 || ucode < 0) {
			errno = EILSEQ;
			return (size_t) -1;
		}

		return unicode_to_utf8(s, len, ucode);
	}

#endif /* __JIS_CONFIG_HAVE_UNICODE */

	if (wide_to_kuten(wc, &table, &code, &code2)) {
#if 0
		fprintf(stderr, "@@@: %04x -> EILSEQ\n",
			(unsigned) (unsigned short) wc);
#endif
		errno = EILSEQ;
		return (size_t) -1;
	}
#if 0
	fprintf(stderr, "@@@: %04x -> (%d, %d, %d)\n",
		(unsigned) (unsigned short) wc,
		(int) table, code, code2);
#endif

	assert(code >= 0 && code <= 255);

	switch(locale) {
	case JIS_LOCALE_ASCII:
		if (table != ASCII) {
			errno = EILSEQ;
			return (size_t) -1;
		}
		if (len < 1) {
			return 0;
		}
		if (s != NULL) {
			s[0] = code;
		}
		return 1;
			
	case JIS_LOCALE_JIS_ROMAN:
		if (table == ASCII) {
			code = ascii_to_roman(code);
		}
		if (code != -1 && table != JROMAN && table != JKATAKANA) {
			errno = EILSEQ;
			return (size_t) -1;
		}
		if (len < 1) {
			return 0;
		}
		if (s != NULL) {
			s[0] = code;
		}
		return 1;
			
	case JIS_LOCALE_JIS:
		if (table == ASCII) {
			code2 = ascii_to_roman(code);
			if (code2 == -1) {
				table = JROMAN;
				code = code2;
			}
		}

		new_state = jis_shifts[table].state;
		bytelen = jis_shifts[table].bytelen;
		if (bytelen == 2 && (code > 94 || code2 > 94)) {
			errno = EILSEQ;
			return (size_t) -1;
		}
		if (bytelen > len) {
			return 0;
		}
		shiftlen = 0;
		if (new_state != state->state) {
			shiftlen = jis_shifts[table].shiftlen;
			if (bytelen + shiftlen > len) {
				return 0;
			}
			state->state = new_state;
			if (s != NULL) {
				memcpy(s, jis_shifts[table].shift, shiftlen);
				s += shiftlen;
			}
		}
		if (s != NULL) {
			if (bytelen == 1) {
				if (table == JKATAKANA) {
					code -= 128;
				}
				s[0] = (char) code;
			} else {
				s[0] = 32 + code;
				s[1] = 32 + code2;
			}
		}
		return shiftlen + bytelen;
		
	case JIS_LOCALE_EUC:
#if 0
		fprintf(stderr, "@@@: euc2 table = %d, code = %ld,%ld\n",
			(int) table, (long) code, (long) code2);
#endif
		switch(table) {
#ifdef CONFIG_EUC_ASCII
 		case JROMAN:
			code = roman_to_ascii(code);
			/* FALLTHROUGH */
		case ASCII:
#else
 		case ASCII:
			code = ascii_to_roman(code);
			/* FALLTHROUGH */
		case JROMAN:
#endif
			if (code == -1 || (code >= 161 && code <= 254)
			    || code == 142 || code == 143) {
				errno = EILSEQ;
				return (size_t) -1;
			}
			if (len < 1) {
				return 0;
			}
			if (s != NULL) {
				s[0] = (char) code;
			}
			return 1;

		case JKATAKANA:
			if (len < 2) {
				return 0;
			}
			if (s != NULL) {
				s[0] = (char) 142;
				s[1] = (char) code;
			}
			return 2;

		case J1990:
 		case J1983:
			if (code > 94 || code2 > 94) {
				errno = EILSEQ;
				return (size_t) -1;
			}
			if (len < 2) {
				return 0;
			}
			if (s != NULL) {
				s[0] = (char) (160 + code);
				s[1] = (char) (160 + code2);
			}
			return 2;
		
 		case J0212:
			if (code > 94 || code2 > 94) {
				errno = EILSEQ;
				return (size_t) -1;
			}
			if (len < 3) {
				return 0;
			}
			if (s != NULL) {
				s[0] = (char) 143;
				s[1] = (char) (160 + code);
				s[2] = (char) (160 + code2);
			}
			return 3;
		default:
			break;
		}
		assert(0);
		break;

	case JIS_LOCALE_SHIFT_JIS:
		switch(table) {
		case ASCII:
			code = ascii_to_roman(code);
			if (code == -1) {
				errno = EILSEQ;
				return (size_t) -1;
			}
			/* FALLTHROUGH */
 		case JROMAN:
		case JKATAKANA:
			if (len < 1) {
				return 0;
			}
			if (s != NULL) {
				s[0] = (char) code;
			}
			return 1;

		case J1983:
		case J1990:
			if (len < 2) {
				return 0;
			}
			if (code > 120 || code2 > 94) {
				errno = EILSEQ;
				return (size_t) -1;
			}
			if (s != NULL) {
				code--;
				if (code & 1) {
					code2 += 63 + 94;
				} else {
					code2 += 63;
				}
				if (code2 > 126) {
					code2++;
				}
				code = (code >> 1) + 129;
				if (code > 159) {
					code += 64;
				}
				assert(code <= 0xFC);
				s[0] = (char) code;
				s[1] = (char) code2;
			}
			return 2;

 		case J0212:
			errno = EILSEQ;
			return (size_t) -1;

		default:
			assert(0);
		}
		break;

	case JIS_LOCALE_QUERY:
	case JIS_LOCALE_DEFAULT:
#ifdef __JIS_CONFIG_HAVE_UNICODE
	case JIS_LOCALE_UTF8:
#endif
		break;
	}
	assert(0);
	abort();
}
